package org.eclipse.viatra.cep.core.engine.runtime;

import java.util.Collection;
import java.util.HashSet;
import java.util.Set;
import org.apache.log4j.Logger;
import org.eclipse.viatra.cep.core.engine.runtime.InitStateMatch;
import org.eclipse.viatra.cep.core.engine.runtime.util.InitStateQuerySpecification;
import org.eclipse.viatra.cep.core.metamodels.automaton.Automaton;
import org.eclipse.viatra.cep.core.metamodels.automaton.InitState;
import org.eclipse.viatra.query.runtime.api.IMatchProcessor;
import org.eclipse.viatra.query.runtime.api.IQuerySpecification;
import org.eclipse.viatra.query.runtime.api.ViatraQueryEngine;
import org.eclipse.viatra.query.runtime.api.impl.BaseMatcher;
import org.eclipse.viatra.query.runtime.exception.ViatraQueryException;
import org.eclipse.viatra.query.runtime.matchers.tuple.Tuple;
import org.eclipse.viatra.query.runtime.util.ViatraQueryLoggingUtil;

/**
 * Generated pattern matcher API of the org.eclipse.viatra.cep.core.engine.runtime.initState pattern,
 * providing pattern-specific query methods.
 * 
 * <p>Use the pattern matcher on a given model via {@link #on(ViatraQueryEngine)},
 * e.g. in conjunction with {@link ViatraQueryEngine#on(Notifier)}.
 * 
 * <p>Matches of the pattern will be represented as {@link InitStateMatch}.
 * 
 * <p>Original source:
 * <code><pre>
 * pattern initState(automaton : Automaton, initState : InitState) {
 * 	Automaton.initialState(automaton, initState);
 * }
 * </pre></code>
 * 
 * @see InitStateMatch
 * @see InitStateProcessor
 * @see InitStateQuerySpecification
 * 
 */
@SuppressWarnings("all")
public class InitStateMatcher extends BaseMatcher<InitStateMatch> {
  /**
   * Initializes the pattern matcher within an existing VIATRA Query engine.
   * If the pattern matcher is already constructed in the engine, only a light-weight reference is returned.
   * The match set will be incrementally refreshed upon updates.
   * @param engine the existing VIATRA Query engine in which this matcher will be created.
   * @throws ViatraQueryException if an error occurs during pattern matcher creation
   * 
   */
  public static InitStateMatcher on(final ViatraQueryEngine engine) throws ViatraQueryException {
    // check if matcher already exists
    InitStateMatcher matcher = engine.getExistingMatcher(querySpecification());
    if (matcher == null) {
    	matcher = new InitStateMatcher(engine);
    	// do not have to "put" it into engine.matchers, reportMatcherInitialized() will take care of it
    }
    return matcher;
  }
  
  private final static int POSITION_AUTOMATON = 0;
  
  private final static int POSITION_INITSTATE = 1;
  
  private final static Logger LOGGER = ViatraQueryLoggingUtil.getLogger(InitStateMatcher.class);
  
  /**
   * Initializes the pattern matcher within an existing VIATRA Query engine.
   * If the pattern matcher is already constructed in the engine, only a light-weight reference is returned.
   * The match set will be incrementally refreshed upon updates.
   * @param engine the existing VIATRA Query engine in which this matcher will be created.
   * @throws ViatraQueryException if an error occurs during pattern matcher creation
   * 
   */
  private InitStateMatcher(final ViatraQueryEngine engine) throws ViatraQueryException {
    super(engine, querySpecification());
  }
  
  /**
   * Returns the set of all matches of the pattern that conform to the given fixed values of some parameters.
   * @param pAutomaton the fixed value of pattern parameter automaton, or null if not bound.
   * @param pInitState the fixed value of pattern parameter initState, or null if not bound.
   * @return matches represented as a InitStateMatch object.
   * 
   */
  public Collection<InitStateMatch> getAllMatches(final Automaton pAutomaton, final InitState pInitState) {
    return rawGetAllMatches(new Object[]{pAutomaton, pInitState});
  }
  
  /**
   * Returns an arbitrarily chosen match of the pattern that conforms to the given fixed values of some parameters.
   * Neither determinism nor randomness of selection is guaranteed.
   * @param pAutomaton the fixed value of pattern parameter automaton, or null if not bound.
   * @param pInitState the fixed value of pattern parameter initState, or null if not bound.
   * @return a match represented as a InitStateMatch object, or null if no match is found.
   * 
   */
  public InitStateMatch getOneArbitraryMatch(final Automaton pAutomaton, final InitState pInitState) {
    return rawGetOneArbitraryMatch(new Object[]{pAutomaton, pInitState});
  }
  
  /**
   * Indicates whether the given combination of specified pattern parameters constitute a valid pattern match,
   * under any possible substitution of the unspecified parameters (if any).
   * @param pAutomaton the fixed value of pattern parameter automaton, or null if not bound.
   * @param pInitState the fixed value of pattern parameter initState, or null if not bound.
   * @return true if the input is a valid (partial) match of the pattern.
   * 
   */
  public boolean hasMatch(final Automaton pAutomaton, final InitState pInitState) {
    return rawHasMatch(new Object[]{pAutomaton, pInitState});
  }
  
  /**
   * Returns the number of all matches of the pattern that conform to the given fixed values of some parameters.
   * @param pAutomaton the fixed value of pattern parameter automaton, or null if not bound.
   * @param pInitState the fixed value of pattern parameter initState, or null if not bound.
   * @return the number of pattern matches found.
   * 
   */
  public int countMatches(final Automaton pAutomaton, final InitState pInitState) {
    return rawCountMatches(new Object[]{pAutomaton, pInitState});
  }
  
  /**
   * Executes the given processor on each match of the pattern that conforms to the given fixed values of some parameters.
   * @param pAutomaton the fixed value of pattern parameter automaton, or null if not bound.
   * @param pInitState the fixed value of pattern parameter initState, or null if not bound.
   * @param processor the action that will process each pattern match.
   * 
   */
  public void forEachMatch(final Automaton pAutomaton, final InitState pInitState, final IMatchProcessor<? super InitStateMatch> processor) {
    rawForEachMatch(new Object[]{pAutomaton, pInitState}, processor);
  }
  
  /**
   * Executes the given processor on an arbitrarily chosen match of the pattern that conforms to the given fixed values of some parameters.
   * Neither determinism nor randomness of selection is guaranteed.
   * @param pAutomaton the fixed value of pattern parameter automaton, or null if not bound.
   * @param pInitState the fixed value of pattern parameter initState, or null if not bound.
   * @param processor the action that will process the selected match.
   * @return true if the pattern has at least one match with the given parameter values, false if the processor was not invoked
   * 
   */
  public boolean forOneArbitraryMatch(final Automaton pAutomaton, final InitState pInitState, final IMatchProcessor<? super InitStateMatch> processor) {
    return rawForOneArbitraryMatch(new Object[]{pAutomaton, pInitState}, processor);
  }
  
  /**
   * Returns a new (partial) match.
   * This can be used e.g. to call the matcher with a partial match.
   * <p>The returned match will be immutable. Use {@link #newEmptyMatch()} to obtain a mutable match object.
   * @param pAutomaton the fixed value of pattern parameter automaton, or null if not bound.
   * @param pInitState the fixed value of pattern parameter initState, or null if not bound.
   * @return the (partial) match object.
   * 
   */
  public InitStateMatch newMatch(final Automaton pAutomaton, final InitState pInitState) {
    return InitStateMatch.newMatch(pAutomaton, pInitState);
  }
  
  /**
   * Retrieve the set of values that occur in matches for automaton.
   * @return the Set of all values, null if no parameter with the given name exists, empty set if there are no matches
   * 
   */
  protected Set<Automaton> rawAccumulateAllValuesOfautomaton(final Object[] parameters) {
    Set<Automaton> results = new HashSet<Automaton>();
    rawAccumulateAllValues(POSITION_AUTOMATON, parameters, results);
    return results;
  }
  
  /**
   * Retrieve the set of values that occur in matches for automaton.
   * @return the Set of all values, null if no parameter with the given name exists, empty set if there are no matches
   * 
   */
  public Set<Automaton> getAllValuesOfautomaton() {
    return rawAccumulateAllValuesOfautomaton(emptyArray());
  }
  
  /**
   * Retrieve the set of values that occur in matches for automaton.
   * @return the Set of all values, null if no parameter with the given name exists, empty set if there are no matches
   * 
   */
  public Set<Automaton> getAllValuesOfautomaton(final InitStateMatch partialMatch) {
    return rawAccumulateAllValuesOfautomaton(partialMatch.toArray());
  }
  
  /**
   * Retrieve the set of values that occur in matches for automaton.
   * @return the Set of all values, null if no parameter with the given name exists, empty set if there are no matches
   * 
   */
  public Set<Automaton> getAllValuesOfautomaton(final InitState pInitState) {
    return rawAccumulateAllValuesOfautomaton(new Object[]{
    null, 
    pInitState
    });
  }
  
  /**
   * Retrieve the set of values that occur in matches for initState.
   * @return the Set of all values, null if no parameter with the given name exists, empty set if there are no matches
   * 
   */
  protected Set<InitState> rawAccumulateAllValuesOfinitState(final Object[] parameters) {
    Set<InitState> results = new HashSet<InitState>();
    rawAccumulateAllValues(POSITION_INITSTATE, parameters, results);
    return results;
  }
  
  /**
   * Retrieve the set of values that occur in matches for initState.
   * @return the Set of all values, null if no parameter with the given name exists, empty set if there are no matches
   * 
   */
  public Set<InitState> getAllValuesOfinitState() {
    return rawAccumulateAllValuesOfinitState(emptyArray());
  }
  
  /**
   * Retrieve the set of values that occur in matches for initState.
   * @return the Set of all values, null if no parameter with the given name exists, empty set if there are no matches
   * 
   */
  public Set<InitState> getAllValuesOfinitState(final InitStateMatch partialMatch) {
    return rawAccumulateAllValuesOfinitState(partialMatch.toArray());
  }
  
  /**
   * Retrieve the set of values that occur in matches for initState.
   * @return the Set of all values, null if no parameter with the given name exists, empty set if there are no matches
   * 
   */
  public Set<InitState> getAllValuesOfinitState(final Automaton pAutomaton) {
    return rawAccumulateAllValuesOfinitState(new Object[]{
    pAutomaton, 
    null
    });
  }
  
  @Override
  protected InitStateMatch tupleToMatch(final Tuple t) {
    try {
    	return InitStateMatch.newMatch((Automaton) t.get(POSITION_AUTOMATON), (InitState) t.get(POSITION_INITSTATE));
    } catch(ClassCastException e) {
    	LOGGER.error("Element(s) in tuple not properly typed!",e);
    	return null;
    }
  }
  
  @Override
  protected InitStateMatch arrayToMatch(final Object[] match) {
    try {
    	return InitStateMatch.newMatch((Automaton) match[POSITION_AUTOMATON], (InitState) match[POSITION_INITSTATE]);
    } catch(ClassCastException e) {
    	LOGGER.error("Element(s) in array not properly typed!",e);
    	return null;
    }
  }
  
  @Override
  protected InitStateMatch arrayToMatchMutable(final Object[] match) {
    try {
    	return InitStateMatch.newMutableMatch((Automaton) match[POSITION_AUTOMATON], (InitState) match[POSITION_INITSTATE]);
    } catch(ClassCastException e) {
    	LOGGER.error("Element(s) in array not properly typed!",e);
    	return null;
    }
  }
  
  /**
   * @return the singleton instance of the query specification of this pattern
   * @throws ViatraQueryException if the pattern definition could not be loaded
   * 
   */
  public static IQuerySpecification<InitStateMatcher> querySpecification() throws ViatraQueryException {
    return InitStateQuerySpecification.instance();
  }
}
