package org.eclipse.viatra.query.tooling.ui.retevis;

import java.util.Collection;
import java.util.HashSet;
import java.util.Set;
import org.apache.log4j.Logger;
import org.eclipse.viatra.query.runtime.api.IMatchProcessor;
import org.eclipse.viatra.query.runtime.api.IQuerySpecification;
import org.eclipse.viatra.query.runtime.api.ViatraQueryEngine;
import org.eclipse.viatra.query.runtime.api.impl.BaseMatcher;
import org.eclipse.viatra.query.runtime.exception.ViatraQueryException;
import org.eclipse.viatra.query.runtime.matchers.tuple.Tuple;
import org.eclipse.viatra.query.runtime.rete.recipes.InputRecipe;
import org.eclipse.viatra.query.runtime.util.ViatraQueryLoggingUtil;
import org.eclipse.viatra.query.tooling.ui.retevis.InputRecipeMatch;
import org.eclipse.viatra.query.tooling.ui.retevis.util.InputRecipeQuerySpecification;

/**
 * Generated pattern matcher API of the org.eclipse.viatra.query.tooling.ui.retevis.inputRecipe pattern,
 * providing pattern-specific query methods.
 * 
 * <p>Use the pattern matcher on a given model via {@link #on(ViatraQueryEngine)},
 * e.g. in conjunction with {@link ViatraQueryEngine#on(Notifier)}.
 * 
 * <p>Matches of the pattern will be represented as {@link InputRecipeMatch}.
 * 
 * <p>Original source:
 * <code><pre>
 * {@literal @}Item(item = recipe, label = "Input: $recipe.inputKey$")
 * {@literal @}Format(textColor = "#FFFFFF", color = "#7F004D")
 * {@literal @}QueryExplorer(checked = false)
 * pattern inputRecipe(recipe : InputRecipe) {
 * 	InputRecipe(recipe);
 * }
 * </pre></code>
 * 
 * @see InputRecipeMatch
 * @see InputRecipeProcessor
 * @see InputRecipeQuerySpecification
 * 
 */
@SuppressWarnings("all")
public class InputRecipeMatcher extends BaseMatcher<InputRecipeMatch> {
  /**
   * Initializes the pattern matcher within an existing VIATRA Query engine.
   * If the pattern matcher is already constructed in the engine, only a light-weight reference is returned.
   * The match set will be incrementally refreshed upon updates.
   * @param engine the existing VIATRA Query engine in which this matcher will be created.
   * @throws ViatraQueryException if an error occurs during pattern matcher creation
   * 
   */
  public static InputRecipeMatcher on(final ViatraQueryEngine engine) throws ViatraQueryException {
    // check if matcher already exists
    InputRecipeMatcher matcher = engine.getExistingMatcher(querySpecification());
    if (matcher == null) {
    	matcher = new InputRecipeMatcher(engine);
    	// do not have to "put" it into engine.matchers, reportMatcherInitialized() will take care of it
    }
    return matcher;
  }
  
  private final static int POSITION_RECIPE = 0;
  
  private final static Logger LOGGER = ViatraQueryLoggingUtil.getLogger(InputRecipeMatcher.class);
  
  /**
   * Initializes the pattern matcher within an existing VIATRA Query engine.
   * If the pattern matcher is already constructed in the engine, only a light-weight reference is returned.
   * The match set will be incrementally refreshed upon updates.
   * @param engine the existing VIATRA Query engine in which this matcher will be created.
   * @throws ViatraQueryException if an error occurs during pattern matcher creation
   * 
   */
  private InputRecipeMatcher(final ViatraQueryEngine engine) throws ViatraQueryException {
    super(engine, querySpecification());
  }
  
  /**
   * Returns the set of all matches of the pattern that conform to the given fixed values of some parameters.
   * @param pRecipe the fixed value of pattern parameter recipe, or null if not bound.
   * @return matches represented as a InputRecipeMatch object.
   * 
   */
  public Collection<InputRecipeMatch> getAllMatches(final InputRecipe pRecipe) {
    return rawGetAllMatches(new Object[]{pRecipe});
  }
  
  /**
   * Returns an arbitrarily chosen match of the pattern that conforms to the given fixed values of some parameters.
   * Neither determinism nor randomness of selection is guaranteed.
   * @param pRecipe the fixed value of pattern parameter recipe, or null if not bound.
   * @return a match represented as a InputRecipeMatch object, or null if no match is found.
   * 
   */
  public InputRecipeMatch getOneArbitraryMatch(final InputRecipe pRecipe) {
    return rawGetOneArbitraryMatch(new Object[]{pRecipe});
  }
  
  /**
   * Indicates whether the given combination of specified pattern parameters constitute a valid pattern match,
   * under any possible substitution of the unspecified parameters (if any).
   * @param pRecipe the fixed value of pattern parameter recipe, or null if not bound.
   * @return true if the input is a valid (partial) match of the pattern.
   * 
   */
  public boolean hasMatch(final InputRecipe pRecipe) {
    return rawHasMatch(new Object[]{pRecipe});
  }
  
  /**
   * Returns the number of all matches of the pattern that conform to the given fixed values of some parameters.
   * @param pRecipe the fixed value of pattern parameter recipe, or null if not bound.
   * @return the number of pattern matches found.
   * 
   */
  public int countMatches(final InputRecipe pRecipe) {
    return rawCountMatches(new Object[]{pRecipe});
  }
  
  /**
   * Executes the given processor on each match of the pattern that conforms to the given fixed values of some parameters.
   * @param pRecipe the fixed value of pattern parameter recipe, or null if not bound.
   * @param processor the action that will process each pattern match.
   * 
   */
  public void forEachMatch(final InputRecipe pRecipe, final IMatchProcessor<? super InputRecipeMatch> processor) {
    rawForEachMatch(new Object[]{pRecipe}, processor);
  }
  
  /**
   * Executes the given processor on an arbitrarily chosen match of the pattern that conforms to the given fixed values of some parameters.
   * Neither determinism nor randomness of selection is guaranteed.
   * @param pRecipe the fixed value of pattern parameter recipe, or null if not bound.
   * @param processor the action that will process the selected match.
   * @return true if the pattern has at least one match with the given parameter values, false if the processor was not invoked
   * 
   */
  public boolean forOneArbitraryMatch(final InputRecipe pRecipe, final IMatchProcessor<? super InputRecipeMatch> processor) {
    return rawForOneArbitraryMatch(new Object[]{pRecipe}, processor);
  }
  
  /**
   * Returns a new (partial) match.
   * This can be used e.g. to call the matcher with a partial match.
   * <p>The returned match will be immutable. Use {@link #newEmptyMatch()} to obtain a mutable match object.
   * @param pRecipe the fixed value of pattern parameter recipe, or null if not bound.
   * @return the (partial) match object.
   * 
   */
  public InputRecipeMatch newMatch(final InputRecipe pRecipe) {
    return InputRecipeMatch.newMatch(pRecipe);
  }
  
  /**
   * Retrieve the set of values that occur in matches for recipe.
   * @return the Set of all values, null if no parameter with the given name exists, empty set if there are no matches
   * 
   */
  protected Set<InputRecipe> rawAccumulateAllValuesOfrecipe(final Object[] parameters) {
    Set<InputRecipe> results = new HashSet<InputRecipe>();
    rawAccumulateAllValues(POSITION_RECIPE, parameters, results);
    return results;
  }
  
  /**
   * Retrieve the set of values that occur in matches for recipe.
   * @return the Set of all values, null if no parameter with the given name exists, empty set if there are no matches
   * 
   */
  public Set<InputRecipe> getAllValuesOfrecipe() {
    return rawAccumulateAllValuesOfrecipe(emptyArray());
  }
  
  @Override
  protected InputRecipeMatch tupleToMatch(final Tuple t) {
    try {
    	return InputRecipeMatch.newMatch((InputRecipe) t.get(POSITION_RECIPE));
    } catch(ClassCastException e) {
    	LOGGER.error("Element(s) in tuple not properly typed!",e);
    	return null;
    }
  }
  
  @Override
  protected InputRecipeMatch arrayToMatch(final Object[] match) {
    try {
    	return InputRecipeMatch.newMatch((InputRecipe) match[POSITION_RECIPE]);
    } catch(ClassCastException e) {
    	LOGGER.error("Element(s) in array not properly typed!",e);
    	return null;
    }
  }
  
  @Override
  protected InputRecipeMatch arrayToMatchMutable(final Object[] match) {
    try {
    	return InputRecipeMatch.newMutableMatch((InputRecipe) match[POSITION_RECIPE]);
    } catch(ClassCastException e) {
    	LOGGER.error("Element(s) in array not properly typed!",e);
    	return null;
    }
  }
  
  /**
   * @return the singleton instance of the query specification of this pattern
   * @throws ViatraQueryException if the pattern definition could not be loaded
   * 
   */
  public static IQuerySpecification<InputRecipeMatcher> querySpecification() throws ViatraQueryException {
    return InputRecipeQuerySpecification.instance();
  }
}
